#include "includes.h"
#include <sys/types.h>
#include <openssl/bn.h>
#include <openssl/ecdsa.h>
#include <openssl/evp.h>

#include <string.h>
#include "sshbuf.h"
#include "ssherr.h"
#include "digest.h"
#include "sshkey.h"

#include "openbsd-compat/openssl-compat.h"

const unsigned char *sm2_id = (const unsigned char *)"1234567812345678";

int
ssh_sm2_sign(const struct sshkey *key, u_char **sigp, size_t *lenp,
    const u_char *data, size_t datalen, u_int compat)
{
	u_char *sig = NULL;
        size_t slen = 0;
	int pkey_len = 0;
	int r = 0;
	int len = 0;
	EVP_PKEY *key_sm2 = NULL;
        struct sshbuf *b = NULL;
	EVP_PKEY_CTX *pctx = NULL;
	EVP_MD_CTX *mctx = NULL;
        int ret = SSH_ERR_INTERNAL_ERROR;

        if (lenp != NULL)
                *lenp = 0;
        if (sigp != NULL)
                *sigp = NULL;

	if (key == NULL || key->ecdsa == NULL ||
	    sshkey_type_plain(key->type) != KEY_SM2)
		return SSH_ERR_INVALID_ARGUMENT;

	if ((key_sm2 = EVP_PKEY_new()) == NULL) {
		return SSH_ERR_ALLOC_FAIL;
	}

	if ((EVP_PKEY_set1_EC_KEY(key_sm2, key->ecdsa)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
		goto out;
	}

        if ((pkey_len = EVP_PKEY_size(key_sm2)) == 0) {
                ret = SSH_ERR_INVALID_ARGUMENT;
		goto out;
        }

	slen = pkey_len;
	
	if ((EVP_PKEY_set_alias_type(key_sm2, EVP_PKEY_SM2)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
		goto out;
	}

        if ((sig = OPENSSL_malloc(pkey_len)) == NULL) {
		ret = SSH_ERR_ALLOC_FAIL;
		goto out;
        }

	if ((pctx = EVP_PKEY_CTX_new(key_sm2, NULL)) == NULL) {
		ret = SSH_ERR_ALLOC_FAIL;
                goto out;
	}

	if (EVP_PKEY_CTX_set1_id(pctx, sm2_id, 16) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
	}
	
	if ((mctx = EVP_MD_CTX_new()) == NULL) {
                ret = SSH_ERR_ALLOC_FAIL;
                goto out;
	}

	EVP_MD_CTX_set_pkey_ctx(mctx, pctx);

        if ((EVP_DigestSignInit(mctx, NULL, EVP_sm3(), NULL, key_sm2)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
		goto out;
	}

        if ((EVP_DigestSignUpdate(mctx, data, datalen)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
	}
        
	if ((EVP_DigestSignFinal(mctx, sig, &slen)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
	}

	if ((b = sshbuf_new()) == NULL) {
                ret = SSH_ERR_ALLOC_FAIL;
                goto out;
        }

        if ((r = sshbuf_put_cstring(b, "sm2")) != 0 ||
            (r = sshbuf_put_string(b, sig, slen)) != 0)
                goto out;
	len = sshbuf_len(b);
	if (sigp != NULL) {
                if ((*sigp = malloc(len)) == NULL) {
                        ret = SSH_ERR_ALLOC_FAIL;
                        goto out;
                }
                memcpy(*sigp, sshbuf_ptr(b), len);
        } 
	if (lenp != NULL)
                *lenp = len;
	ret = 0;

out:
	EVP_PKEY_free(key_sm2);
        if (sig != NULL) {
                explicit_bzero(sig, slen);
                OPENSSL_free(sig);
        }
	EVP_PKEY_CTX_free(pctx);
	EVP_MD_CTX_free(mctx);
        sshbuf_free(b);
	return ret;
}

int
ssh_sm2_verify(const struct sshkey *key,
    const u_char *signature, size_t signaturelen,
    const u_char *data, size_t datalen, u_int compat)
{
	const u_char *sig = NULL;
	char *ktype = NULL;
	size_t slen = 0;
        int pkey_len = 0;
        int r = 0;
        int len = 0;
	EVP_PKEY *key_sm2 = NULL;
        struct sshbuf *b = NULL;
        EVP_PKEY_CTX *pctx = NULL;
        EVP_MD_CTX *mctx = NULL;
        int ret = SSH_ERR_INTERNAL_ERROR;

        if (key == NULL ||
            sshkey_type_plain(key->type) != KEY_SM2 ||
            signature == NULL || signaturelen == 0)
                return SSH_ERR_INVALID_ARGUMENT;

        if ((b = sshbuf_from(signature, signaturelen)) == NULL)
                return SSH_ERR_ALLOC_FAIL;

        if ((r = sshbuf_get_cstring(b, &ktype, NULL)) != 0 ||
            (r = sshbuf_get_string_direct(b, &sig, &slen)) != 0)
                goto out;

	if (strcmp("sm2", ktype) != 0) {
                ret = SSH_ERR_KEY_TYPE_MISMATCH;
                goto out;
	}

        if (sshbuf_len(b) != 0) {
                ret = SSH_ERR_UNEXPECTED_TRAILING_DATA;
                goto out;
        }

        if ((key_sm2 = EVP_PKEY_new()) == NULL) {
                ret = SSH_ERR_ALLOC_FAIL;
                goto out;
        }

        if ((EVP_PKEY_set1_EC_KEY(key_sm2, key->ecdsa)) != 1) {
                ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
        }

        if ((pkey_len = EVP_PKEY_size(key_sm2)) == 0) {
                ret = SSH_ERR_INVALID_ARGUMENT;
                goto out;
        }

        if ((EVP_PKEY_set_alias_type(key_sm2, EVP_PKEY_SM2)) != 1) {
                ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
        }

	if ((pctx = EVP_PKEY_CTX_new(key_sm2, NULL)) == NULL) {
                ret = SSH_ERR_ALLOC_FAIL;
                goto out;
        }

        if (EVP_PKEY_CTX_set1_id(pctx, sm2_id, 16) != 1) {
                ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
        }

        if ((mctx = EVP_MD_CTX_new()) == NULL) {
                ret = SSH_ERR_ALLOC_FAIL;
                goto out;
        }

	EVP_MD_CTX_set_pkey_ctx(mctx, pctx);

        if ((EVP_DigestVerifyInit(mctx, NULL, EVP_sm3(), NULL, key_sm2)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
	}
        
	if ((EVP_DigestVerifyUpdate(mctx, data, datalen)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
	}
        
	if ((EVP_DigestVerifyFinal(mctx, sig, slen)) != 1) {
		ret = SSH_ERR_INTERNAL_ERROR;
                goto out;
	}

	ret = 0;
out:
	EVP_PKEY_free(key_sm2);
        EVP_PKEY_CTX_free(pctx);
        EVP_MD_CTX_free(mctx);
        sshbuf_free(b);
	free(ktype);
	return ret;
}
